"""BLE device abstraction for SOK batteries."""

from __future__ import annotations

import asyncio
import logging
import statistics
import struct
from contextlib import asynccontextmanager
from typing import AsyncIterator, Optional

import async_timeout
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.exc import BleakError

from sok_ble.const import UUID_RX, UUID_TX, _sok_command
from sok_ble.exceptions import BLEConnectionError
from sok_ble.sok_parser import SokParser

logger = logging.getLogger(__name__)

try:
    from bleak_retry_connector import (
        BleakClientWithServiceCache,
        establish_connection,
    )
except Exception:  # pragma: no cover - optional dependency
    from bleak import BleakClient as BleakClientWithServiceCache

    establish_connection = None  # type: ignore[misc]


class SokBluetoothDevice:
    """Minimal BLE interface for a SOK battery."""

    def __init__(
        self, ble_device: BLEDevice, adapter: Optional[str] | None = None
    ) -> None:
        self._ble_device = ble_device
        self._adapter = adapter

        self.voltage: float | None = None
        self.current: float | None = None
        self.soc: int | None = None
        self.temperature: float | None = None
        self.capacity: float | None = None
        self.num_cycles: int | None = None
        self.cell_voltages: list[float] | None = None

        # Housekeeping
        self.num_samples = 0

    @asynccontextmanager
    async def _connect(self) -> AsyncIterator[BleakClientWithServiceCache]:
        """Connect to the device and yield a BLE client."""
        logger.debug("Connecting to %s", self._ble_device.address)
        last_err: Exception | None = None
        client: BleakClientWithServiceCache | None = None

        for attempt in range(3):
            try:
                if establish_connection:
                    client = await establish_connection(
                        BleakClientWithServiceCache,
                        self._ble_device,
                        self._ble_device.name or self._ble_device.address,
                        adapter=self._adapter,
                    )
                else:
                    client = BleakClientWithServiceCache(
                        self._ble_device,
                        adapter=self._adapter,
                    )
                    await client.connect()

                # Force service discovery
                async with async_timeout.timeout(5):
                    _ = client.services
                await asyncio.sleep(0.15)
                break
            except (BleakError, asyncio.TimeoutError) as err:
                last_err = err
                logger.debug(
                    "BLE connect attempt %s failed for %s: %s",
                    attempt + 1,
                    self._ble_device.address,
                    err,
                )
                await asyncio.sleep(0.5)
        else:
            raise BLEConnectionError(
                f"Unable to establish GATT connection to {self._ble_device.address}"
            ) from last_err

        assert client is not None
        try:
            yield client
        finally:
            await client.disconnect()
            logger.debug("Disconnected from %s", self._ble_device.address)

    async def _send_command(
        self, client: BleakClientWithServiceCache, cmd: int, expected: int
    ) -> bytes:
        """Send a command and return the response bytes with the given header."""

        for attempt in range(2):
            try:
                start_notify = getattr(client, "start_notify", None)
                if start_notify is None:
                    await client.write_gatt_char(UUID_TX, _sok_command(cmd))
                    data = bytes(await client.read_gatt_char(UUID_RX))
                    return data

                queue: asyncio.Queue[bytes] = asyncio.Queue()

                def handler(_: BleakGATTCharacteristic, data: bytearray) -> None:
                    queue.put_nowait(bytes(data))

                await client.start_notify(UUID_RX, handler)
                try:
                    await client.write_gatt_char(UUID_TX, _sok_command(cmd))
                    while True:
                        data = await asyncio.wait_for(queue.get(), 5.0)
                        if struct.unpack_from(">H", data)[0] == expected:
                            return data
                finally:
                    await client.stop_notify(UUID_RX)
            except BleakError as err:
                if attempt == 0:
                    logger.debug(
                        "BLE command attempt failed for %s: %s",
                        self._ble_device.address,
                        err,
                    )
                    await asyncio.sleep(0.2)
                    continue
                raise
        raise BleakError(
            f"Failed to receive response 0x{expected:04X} "
            f"from {self._ble_device.address}"
        )

    async def async_update(self) -> None:
        """Poll the device for all telemetry and update attributes."""
        responses: dict[int, bytes] = {}
        async with self._connect() as client:
            logger.debug("Send C1")
            data = await self._send_command(client, 0xC1, 0xCCF0)
            logger.debug(
                "Recv 0x%04X: %s",
                struct.unpack_from(">H", data)[0],
                data.hex(),
            )
            responses[0xCCF0] = data

            logger.debug("Send C1")
            data = await self._send_command(client, 0xC1, 0xCCF2)
            logger.debug(
                "Recv 0x%04X: %s",
                struct.unpack_from(">H", data)[0],
                data.hex(),
            )
            responses[0xCCF2] = data

            logger.debug("Send C2")
            data = await self._send_command(client, 0xC2, 0xCCF3)
            logger.debug(
                "Recv 0x%04X: %s",
                struct.unpack_from(">H", data)[0],
                data.hex(),
            )
            responses[0xCCF3] = data

            logger.debug("Send C2")
            data = await self._send_command(client, 0xC2, 0xCCF4)
            logger.debug(
                "Recv 0x%04X: %s",
                struct.unpack_from(">H", data)[0],
                data.hex(),
            )
            responses[0xCCF4] = data

        parsed = SokParser.parse_all(responses)
        logger.debug("Parsed update: %s", parsed)

        voltage = parsed.get("voltage")
        self.voltage = voltage if isinstance(voltage, (int, float)) else None

        current = parsed.get("current")
        self.current = current if isinstance(current, (int, float)) else None

        soc = parsed.get("soc")
        self.soc = soc if isinstance(soc, int) else None

        temperature = parsed.get("temperature")
        self.temperature = (
            temperature if isinstance(temperature, (int, float)) else None
        )

        capacity = parsed.get("capacity")
        self.capacity = capacity if isinstance(capacity, (int, float)) else None

        num_cycles = parsed.get("num_cycles")
        self.num_cycles = num_cycles if isinstance(num_cycles, int) else None

        cell_voltages = parsed.get("cell_voltages")
        self.cell_voltages = (
            list(cell_voltages) if isinstance(cell_voltages, list) else None
        )

        self.num_samples += 1

    # Derived metrics -----------------------------------------------------

    @property
    def power(self) -> float | None:
        """Return instantaneous power in watts."""
        if self.voltage is None or self.current is None:
            return None
        return self.voltage * self.current

    @property
    def cell_voltage_max(self) -> float | None:
        cells = self.cell_voltages
        return max(cells) if cells else None

    @property
    def cell_voltage_min(self) -> float | None:
        cells = self.cell_voltages
        return min(cells) if cells else None

    @property
    def cell_voltage_avg(self) -> float | None:
        cells = self.cell_voltages
        if not cells:
            return None
        return sum(cells) / len(cells)

    @property
    def cell_voltage_median(self) -> float | None:
        cells = self.cell_voltages
        if not cells:
            return None
        return statistics.median(cells)

    @property
    def cell_voltage_delta(self) -> float | None:
        if self.cell_voltage_max is None or self.cell_voltage_min is None:
            return None
        return self.cell_voltage_max - self.cell_voltage_min

    @property
    def cell_index_max(self) -> int | None:
        cells = self.cell_voltages
        if not cells:
            return None
        return cells.index(max(cells))

    @property
    def cell_index_min(self) -> int | None:
        cells = self.cell_voltages
        if not cells:
            return None
        return cells.index(min(cells))



import pytest
from bleak.backends.device import BLEDevice

from sok_ble.sok_bluetooth_device import SokBluetoothDevice


def make_device():
    return SokBluetoothDevice(BLEDevice("00:11:22:33:44:55", "Test", None, -60))


def test_power_property():
    dev = make_device()
    dev.voltage = 12.5
    dev.current = -10.0
    assert dev.power == pytest.approx(-125.0)


def test_cell_voltage_stats():
    dev = make_device()
    dev.cell_voltages = [3.1, 3.15, 3.05, 3.2]

    assert dev.cell_voltage_max == 3.2
    assert dev.cell_voltage_min == 3.05
    assert dev.cell_voltage_avg == pytest.approx(3.125)
    assert dev.cell_voltage_median == pytest.approx(3.125)
    assert dev.cell_voltage_delta == pytest.approx(0.15)
    assert dev.cell_index_max == 3
    assert dev.cell_index_min == 2



class SokError(Exception):
    """Base for SOK library errors."""


class BLEConnectionError(SokError):
    """Raised when BLE communication fails."""


class InvalidResponseError(SokError):
    """Raised when an invalid response is received."""



import pytest

from sok_ble.exceptions import BLEConnectionError, InvalidResponseError, SokError


def test_exception_inheritance():
    assert issubclass(BLEConnectionError, SokError)
    assert issubclass(InvalidResponseError, SokError)


def test_raises_ble_connection_error():
    with pytest.raises(BLEConnectionError):
        raise BLEConnectionError()


def test_raises_invalid_response_error():
    with pytest.raises(InvalidResponseError):
        raise InvalidResponseError()



import pytest

from sok_ble.sok_parser import SokParser


def test_parse_all():
    info_buf = bytes.fromhex("ccf0000000102700000000000000320041000000")
    temp_buf = bytes.fromhex("ccf2000000140000000000000000000000000000")
    cap_buf = bytes.fromhex("ccf3000000003200000000000000000000000000")
    cell_buf = bytes.fromhex("ccf401c50c0002c60c0003bf0c0004c00c000000")

    responses = {
        0xCCF0: info_buf,
        0xCCF2: temp_buf,
        0xCCF3: cap_buf,
        0xCCF4: cell_buf,
    }

    result = SokParser.parse_all(responses)
    assert result == {
        "voltage": pytest.approx(13.066, rel=1e-3),
        "current": 10.0,
        "soc": 65,
        "temperature": 20.0,
        "capacity": 100.0,
        "num_cycles": 50,
        "cell_voltages": [3.269, 3.27, 3.263, 3.264],
    }



import pytest
from bleak.backends.device import BLEDevice

from sok_ble import sok_bluetooth_device as device_mod


class DummyClient:
    def __init__(self, responses):
        self._responses = list(responses)

    async def connect(self):
        return True

    async def disconnect(self):
        return True

    async def write_gatt_char(self, *args, **kwargs):
        return True

    async def read_gatt_char(self, *args, **kwargs):
        return self._responses.pop(0)

    @property
    def services(self):
        return []


@pytest.mark.asyncio
async def test_full_update(monkeypatch):
    responses = [
        bytes.fromhex("ccf0000000102700000000000000320041000000"),
        bytes.fromhex("ccf2000000140000000000000000000000000000"),
        bytes.fromhex("ccf3000000003200000000000000000000000000"),
        bytes.fromhex("ccf401c50c0002c60c0003bf0c0004c00c000000"),
    ]

    dummy = DummyClient(responses)
    monkeypatch.setattr(device_mod, "establish_connection", None, raising=False)
    monkeypatch.setattr(
        device_mod, "BleakClientWithServiceCache", lambda *a, **k: dummy
    )

    dev = device_mod.SokBluetoothDevice(
        BLEDevice("00:11:22:33:44:55", "Test", None, -60)
    )

    await dev.async_update()

    assert dev.voltage == pytest.approx(13.066)
    assert dev.current == 10.0
    assert dev.soc == 65
    assert dev.temperature == 20.0
    assert dev.capacity == 100.0
    assert dev.num_cycles == 50
    assert dev.cell_voltages == [3.269, 3.27, 3.263, 3.264]
    assert dev.power == pytest.approx(130.66)
    assert dev.cell_voltage_max == 3.27
    assert dev.cell_voltage_min == 3.263
    assert dev.cell_voltage_avg == pytest.approx(3.2665)
    assert dev.cell_voltage_median == pytest.approx(3.2665)
    assert dev.cell_voltage_delta == pytest.approx(0.007)
    assert dev.cell_index_max == 1
    assert dev.cell_index_min == 2



from sok_ble.const import _sok_command, minicrc


def test_minicrc_known_value():
    data = [0xEE, 0xC1, 0x00, 0x00, 0x00]
    assert minicrc(data) == 0xCE


def test_sok_command_crc_and_length():
    cmd = _sok_command(0xC1)
    assert len(cmd) == 6
    # CRC byte should match minicrc of base command frame
    assert cmd[-1] == minicrc([0xEE, 0xC1, 0x00, 0x00, 0x00])



import pytest
from bleak.backends.device import BLEDevice

from sok_ble import sok_bluetooth_device as device_mod


class DummyClient:
    def __init__(self, *args, **kwargs):
        self._responses = [
            bytes.fromhex("ccf0000000102700000000000000320041000000"),
            bytes.fromhex("ccf2000000140000000000000000000000000000"),
            bytes.fromhex("ccf3000000003200000000000000000000000000"),
            bytes.fromhex("ccf401c50c0002c60c0003bf0c0004c00c000000"),
        ]

    async def connect(self):
        return True

    async def disconnect(self):
        return True

    async def write_gatt_char(self, *args, **kwargs):
        return True

    async def read_gatt_char(self, *args, **kwargs):
        return self._responses.pop(0)

    @property
    def services(self):
        return []


@pytest.mark.asyncio
async def test_minimal_update(monkeypatch):
    monkeypatch.setattr(device_mod, "establish_connection", None, raising=False)
    monkeypatch.setattr(device_mod, "BleakClientWithServiceCache", DummyClient)

    dev = device_mod.SokBluetoothDevice(
        BLEDevice("00:11:22:33:44:55", "Test", None, -60)
    )

    await dev.async_update()

    assert dev.voltage == pytest.approx(13.066)
    assert dev.current == 10.0
    assert dev.soc == 65



from sok_ble.sok_parser import SokParser


def test_parse_info_basic():
    hex_data = bytes.fromhex("ccf0000000102700000000000000320041000000")
    result = SokParser.parse_info(hex_data)
    assert result == {
        "current": 10.0,
        "soc": 65,
        "num_cycles": 50,
    }


def test_parse_info_invalid_length():
    data = b"\x00" * 10
    try:
        SokParser.parse_info(data)
        assert False, "Expected InvalidResponseError"
    except Exception as err:
        from sok_ble.exceptions import InvalidResponseError

        assert isinstance(err, InvalidResponseError)



"""Constants and helpers for interacting with SOK BLE batteries."""

from __future__ import annotations

UUID_RX = "0000ffe1-0000-1000-8000-00805f9b34fb"
UUID_TX = "0000ffe2-0000-1000-8000-00805f9b34fb"

# Command byte templates extracted from the reference addon
CMD_NAME = [0xEE, 0xC0, 0x00, 0x00, 0x00]
CMD_INFO = [0xEE, 0xC1, 0x00, 0x00, 0x00]
CMD_DETAIL = [0xEE, 0xC2, 0x00, 0x00, 0x00]
CMD_SETTING = [0xEE, 0xC3, 0x00, 0x00, 0x00]
CMD_PROTECTION = [0xEE, 0xC4, 0x00, 0x00, 0x00]
CMD_BREAK = [0xDD, 0xC0, 0x00, 0x00, 0x00]


def minicrc(data: list[int] | bytes) -> int:
    """Compute CRC-8 used by the SOK protocol."""

    crc = 0
    for byte in data:
        crc ^= byte & 0xFF
        for _ in range(8):
            crc = (crc >> 1) ^ 0x8C if (crc & 1) else crc >> 1
    return crc


def _sok_command(cmd: int) -> bytes:
    """Return a command frame with CRC for the given command byte."""

    data = [0xEE, cmd, 0x00, 0x00, 0x00]
    crc = minicrc(data)
    return bytes(data + [crc])



"""Parsing utilities for SOK BLE responses."""

from __future__ import annotations

import logging
import statistics
import struct
from typing import Dict, Sequence

from sok_ble.exceptions import InvalidResponseError

logger = logging.getLogger(__name__)


# Endian helper functions copied from the reference addon


def get_le_short(data: Sequence[int] | bytes | bytearray, offset: int) -> int:
    """Read a little-endian signed short."""
    return struct.unpack_from("<h", bytes(data), offset)[0]


def get_le_ushort(data: Sequence[int] | bytes | bytearray, offset: int) -> int:
    """Read a little-endian unsigned short."""
    return struct.unpack_from("<H", bytes(data), offset)[0]


def get_le_int3(data: Sequence[int] | bytes | bytearray, offset: int) -> int:
    """Read a 3-byte little-endian signed integer."""
    b0, b1, b2 = bytes(data)[offset : offset + 3]
    val = b0 | (b1 << 8) | (b2 << 16)
    if val & 0x800000:
        val -= 0x1000000
    return val


def get_be_uint3(data: Sequence[int] | bytes | bytearray, offset: int) -> int:
    """Read a 3-byte big-endian unsigned integer."""
    b0, b1, b2 = bytes(data)[offset : offset + 3]
    return (b0 << 16) | (b1 << 8) | b2


class SokParser:
    """Parse buffers returned from SOK batteries."""

    @staticmethod
    def parse_info(buf: bytes) -> Dict[str, float | int]:
        """Parse the information frame for current, SOC and cycles."""
        logger.debug("parse_info input: %s", buf.hex())
        if len(buf) < 20:
            raise InvalidResponseError("Info buffer too short")

        current = get_le_int3(buf, 5) / 1000
        num_cycles = get_le_ushort(buf, 14)
        soc = get_le_ushort(buf, 16)

        result = {
            "current": current,
            "soc": soc,
            "num_cycles": num_cycles,
        }
        logger.debug("parse_info result: %s", result)
        return result

    @staticmethod
    def parse_temps(buf: bytes) -> float:
        """Parse the temperature from the temperature frame."""
        logger.debug("parse_temps input: %s", buf.hex())
        if len(buf) < 20:
            raise InvalidResponseError("Temp buffer too short")

        temperature = get_le_short(buf, 5)
        logger.debug("parse_temps result: %s", temperature)
        return temperature

    @staticmethod
    def parse_capacity_cycles(buf: bytes) -> Dict[str, float | int]:
        """Parse rated capacity."""
        logger.debug("parse_capacity_cycles input: %s", buf.hex())
        if len(buf) < 20:
            raise InvalidResponseError("Capacity buffer too short")

        capacity = get_be_uint3(buf, 5) / 128

        result = {"capacity": capacity}
        logger.debug("parse_capacity_cycles result: %s", result)
        return result

    @staticmethod
    def parse_cells(buf: bytes) -> list[float]:
        """Parse individual cell voltages."""
        logger.debug("parse_cells input: %s", buf.hex())
        if len(buf) < 20:
            raise InvalidResponseError("Cells buffer too short")

        cells = [0.0, 0.0, 0.0, 0.0]
        for x in range(4):
            cell_idx = buf[2 + x * 4]
            cells[cell_idx - 1] = get_le_ushort(buf, 3 + x * 4) / 1000
        logger.debug("parse_cells result: %s", cells)
        return cells

    @classmethod
    def parse_all(
        cls, responses: Dict[int, bytes]
    ) -> Dict[str, float | int | list[float]]:
        """Parse all response buffers into a single dictionary."""
        logger.debug("parse_all input keys: %s", list(responses))
        required = {0xCCF0, 0xCCF2, 0xCCF3, 0xCCF4}
        if not required.issubset(responses):
            raise InvalidResponseError("Missing response buffers")

        info = cls.parse_info(responses[0xCCF0])
        temperature = cls.parse_temps(responses[0xCCF2])
        capacity_info = cls.parse_capacity_cycles(responses[0xCCF3])
        cells = cls.parse_cells(responses[0xCCF4])

        voltage = statistics.mean(cells) * 4

        result = {
            "voltage": voltage,
            "current": info["current"],
            "soc": info["soc"],
            "temperature": temperature,
            "capacity": capacity_info["capacity"],
            "num_cycles": info["num_cycles"],
            "cell_voltages": cells,
        }
        logger.debug("parse_all result: %s", result)
        return result
